/*--------------------------------------------------------------------------+
$Id: codetemplates.xml 1 2011-01-01 00:00:01Z hoelzl $
|                                                                          |
| Copyright 2017 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.af3.platform.raspberry.generator.executable;

import static org.fortiss.af3.expression.utils.ExpressionModelElementFactory.funcCall;
import static org.fortiss.af3.generator.common.utils.CLanguageModelElementFacade.addUserHeaderInclude;
import static org.fortiss.af3.generator.common.utils.SourceModelElementFactory.createByteContentUnitForPluginFile;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.List;
import java.util.Set;

import org.conqat.lib.commons.collections.Pair;
import org.fortiss.af3.component.model.Component;
import org.fortiss.af3.component.model.InputPort;
import org.fortiss.af3.component.model.OutputPort;
import org.fortiss.af3.component.model.Port;
import org.fortiss.af3.component.model.generator.ComponentFunction;
import org.fortiss.af3.component.model.generator.LocalFunction;
import org.fortiss.af3.expression.model.terms.IExpressionTerm;
import org.fortiss.af3.expression.model.terms.imperative.IStatementTerm;
import org.fortiss.af3.generator.common.model.c.CSourcePackage;
import org.fortiss.af3.generator.common.model.source.SourceUnit;
import org.fortiss.af3.platform.generic.generator.executable.GenericExecutionUnitExecutable;
import org.fortiss.af3.platform.language.executable.IInitializableExecutable;
import org.fortiss.af3.platform.language.executable.IReadableExecutable;
import org.fortiss.af3.platform.language.executable.ITerminatableExecutable;
import org.fortiss.af3.platform.language.executable.IWritableExecutable;
import org.fortiss.af3.platform.model.ExecutionUnit;
import org.fortiss.af3.platform.model.PlatformConnectorUnit;
import org.fortiss.af3.platform.model.generic.GenericTransceiver;
import org.fortiss.af3.platform.raspberry.AF3PlatformRaspberryActivator;
import org.fortiss.af3.platform.raspberry.model.RaspberryPi;
import org.fortiss.tooling.kernel.extension.data.ITransformationContext;

/**
 * {@link IInitializableExecutable} and {@link ITerminatableExecutable} for the {@link RaspberryPi}
 * platform.
 * 
 * 
 * @author eder
 * @author $Author: eder $
 * @version $Rev: 18709 $
 * @ConQAT.Rating RED Hash:
 */
public class RaspberryPIExecutable extends GenericExecutionUnitExecutable {

	private static final String SRC_LIB_SUB_PACKAGE_NAME = "src-lib";
	private static final String INC_LIB_SUB_PACKAGE_NAME = "inc-lib";

	/** Constructor. */
	public RaspberryPIExecutable(ExecutionUnit modelElement) {
		super(modelElement);
	}

	/** {@inheritDoc} */
	@Override
	protected CSourcePackage createExecutionUnitSourcePackage(String name,
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts, ITransformationContext context) {

		CSourcePackage pkg =
				super.createExecutionUnitSourcePackage(name, deployedComponents, deployedPorts,
						context);

		// get system.h to add references to self-defined inc files
		SourceUnit system = pkg.getSrcGenPackage().findSourceUnitByName("system.c");

		// Add source/header files here to be copied into the deployment directory
		try {
			/** COMPONENTS **/

			addFile(pkg, "src/main.c");

			// reference is added by GenericExecutableBase -> only add file
			addFile(pkg, "inc/io.h");
			addFile(pkg, "src/io.c");

			// add all components including references
			addFilesWithReference(pkg, system, "CanConnector");
			addFilesWithReference(pkg, system, "Gamepad");
			addFilesWithReference(pkg, system, "GPIO");
			addFilesWithReference(pkg, system, "IMU");
			addFilesWithReference(pkg, system, "LaserRangeFinder");
			addFilesWithReference(pkg, system, "LaserScanner");
			addFilesWithReference(pkg, system, "PWM");

			/** LIBRARIES **/

			// add UART lib
			addFile(pkg, "inc/libuart.h");
			addFile(pkg, "src/libuart.c");
			addFile(pkg, "inc/libmaestro.h");
			addFile(pkg, "src/libmaestro.c");
			addFile(pkg, "inc/libcansocket.h");
			addFile(pkg, "src/libcansocket.c");

			/** FIXES AND PATCHES **/

			// overwrite data_dictionary, the new one has all functions
			// required for the simulation commented out so there are no
			// conflicts
			addFile(pkg, "inc-gen/data_dictionary.h");
			addFile(pkg, "src-gen/data_dictionary.c");

			// overwrite Makedefs, the new one uses -std=gnu99 instead of -std=c99
			// so libuart compiles properly
			addFile(pkg, "Makedefs");

		} catch(IOException | URISyntaxException e) {
			throw new RuntimeException("could not add header and source files from lib", e);
		}

		return pkg;
	}

	private static void addFilesWithReference(CSourcePackage pkg, SourceUnit source,
			String fileWithoutExtension) throws IOException, URISyntaxException {
		addFile(pkg, "src/" + fileWithoutExtension + ".c");
		addFile(pkg, "inc/" + fileWithoutExtension + ".h");
		addSystemCReference(source, fileWithoutExtension);
	}

	private static void addFile(CSourcePackage pkg, String file) throws IOException,
			URISyntaxException {
		pkg.addUnit(createByteContentUnitForPluginFile(AF3PlatformRaspberryActivator.PLUGIN_ID,
				"lib", file, false));
	}

	private static void addSystemCReference(SourceUnit source, String reference) {
		addUserHeaderInclude(source, reference);
	}

	/** {@inheritDoc} */
	@Override
	protected ComponentFunction createInitializeFunction(
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			Set<PlatformConnectorUnit> usedUnits) {
		return super.createInitializeFunction(deployedComponents, usedUnits);
	}

	/** {@inheritDoc} */
	@Override
	protected void
			createReadAccess(IReadableExecutable exec, InputPort p, List<IStatementTerm> body) {
		super.createReadAccess(exec, p, body);
	}

	/** {@inheritDoc} */
	@Override
	protected LocalFunction createReadInputFunction(
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts,
			Set<GenericTransceiver> usedTransceivers) {
		return super.createReadInputFunction(deployedPorts, usedTransceivers);
	}

	/** {@inheritDoc} */
	@Override
	protected ComponentFunction createStepFunction(
			List<Pair<ExecutionUnit, Component>> deployedComponents,
			List<Pair<PlatformConnectorUnit, Port>> deployedPorts,
			Set<GenericTransceiver> usedTransceivers) {
		return super.createStepFunction(deployedComponents, deployedPorts, usedTransceivers);
	}

	/** {@inheritDoc} */
	@Override
	protected LocalFunction createTerminationFunction(Set<PlatformConnectorUnit> usedUnits) {
		return super.createTerminationFunction(usedUnits);
	}

	/** {@inheritDoc} */
	@Override
	protected void createWriteAccess(IWritableExecutable exec, OutputPort p,
			List<IStatementTerm> body) {
		super.createWriteAccess(exec, p, body);
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getInitialization() {
		return funcCall("init_raspberryPI_ecu");
	}

	/** {@inheritDoc} */
	@Override
	public IExpressionTerm getTermination() {
		return funcCall("term_raspberryPI_ecu");
	}

}
